"""
NimbleBox LMAO is our general purpose observability tool for any kind of computation you might have.
"""

import os
import re
from git import Repo
from json import dumps
from requests import Session
from typing import Dict, Any, List, Optional, Union, Tuple
from datetime import datetime, timezone

import nbox.utils as U
from nbox import Instance
from nbox.utils import logger
from nbox.auth import secret
from nbox.nbxlib.tracer import Tracer
from nbox.relics import RelicsNBX
from nbox.sublime.lmao_client import (
  LMAO_Stub, Record, File, FileList, AgentDetails, RunLog, Run, InitRunRequest, ListProjectsRequest, RelicFile
)
from nbox.observability.system import SystemMetricsLogger

DEBUG_LOG_EVERY = 100
INFO_LOG_EVERY = 100


class Lmao():
  def __init__(
    self,
    workspace_id: str,
    project_name: Optional[str] = "",
    metadata: Dict[str, Any] = {},
    project_id: Optional[str] = "",
    save_to_relic: bool = False,
    enable_system_monitoring: bool = False,
  ) -> None:
    """``Lmao`` is the client library for using NimbleBox Monitoring. It talks to your monitoring instance running on your build
    and stores the information in the ``project_name`` or ``project_id``. This object inherently doesn't care what you are actually
    logging and rather concerns itself with ensuring storage."""
    self.workspace_id = workspace_id
    self.save_to_relic = save_to_relic
    self.enable_system_monitoring = enable_system_monitoring

    self.nbx_job_folder = U.env.NBOX_JOB_FOLDER("")
    self._total_logged_elements = 0 # this variable keeps track for logging
    self.completed = False
    self.tracer = None
    self.username = secret.get("username")
    self.relic: RelicsNBX = None
    self.system_monitoring: SystemMetricsLogger = None
    self._nbx_run_id = None
    self._nbx_job_id = None

    # create connection and initialise the run
    self._create_connection(workspace_id)
    self._init(project_name, project_id, config = metadata)

  def __del__(self):
    if self.system_monitoring is not None:
      self.system_monitoring.stop()

  def _create_connection(self, workspace_id: str):
    # prepare the URL
    # id_or_name, workspace_id = U.split_iw(instance_id)
    id_or_name = f"monitoring-{workspace_id}"
    logger.info(f"id_or_name: {id_or_name}")
    logger.info(f"workspace_id: {workspace_id}")
    instance = Instance(id_or_name, workspace_id = workspace_id)
    try:
      open_data = instance.open_data
    except AttributeError:
      raise Exception(f"Is instance '{instance.project_id}' running?")
    build = "build"
    if "app.c." in secret.get("nbx_url"):
      build = "build.c"
    url = f"https://server-{open_data['url']}.{build}.nimblebox.ai/"
    logger.info(f"URL: {url}")

    # create a tracer object that will load all the information
    self.tracer = Tracer(start_heartbeat=False)
    self._nbx_run_id = self.tracer.run_id
    self._nbx_job_id = self.tracer.job_id
    if self._nbx_run_id is not None:
      # update the username you have
      secret.put("username", self.tracer.job_proto.auth_info.username)
      self.username = secret.get("username")

    # create a session with the auth header
    _session = Session()
    _session.headers.update({
      "NBX-TOKEN": open_data["token"],
      "X-NBX-USERNAME": self.username,
    })

    # define the stub
    # self.lmao = LMAO_Stub(url = "http://127.0.0.1:8080", session = Session()) # debug
    self.lmao = LMAO_Stub(url = url, session = _session)

  def _get_name_id(self, project_id: str = None, project_name: str = None):
    all_projects = self.lmao.list_projects(
      _ListProjectsRequest = ListProjectsRequest(
        workspace_id = self.workspace_id,
        project_id_or_name = project_id if project_id else project_name,
      )
    )
    if not all_projects.projects:
      # means we have to initialise this project first and then we will pull the details after the init_run
      if project_id:
        raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
      logger.info(f"Project '{project_name}' not found, will create a new one.")
    
    if project_id:
      common = list(filter(lambda x: x.project_id == project_name, all_projects.projects))
      if len(common) > 1:
        logger.error(f"Multiple entries for {project_id} found, something went wrong from our side.")
        raise Exception(f"Database duplicate entry for project_id: {project_id}")
      elif len(common) == 1:
        project_name = common[0].project_name
        project_id = common[0].project_id
    else:
      common = list(filter(lambda x: x.project_name == project_name, all_projects.projects))
      if len(common) > 1:
        logger.error(f"Project '{project_name}' found multiple times, please use project_id instead.")
        raise Exception("Ambiguous project name, please use project_id instead.")
      elif len(common) == 1:
        project_name = common[0].project_name
        project_id = common[0].project_id
    return project_id, project_name

  def _init(self, project_name, project_id, config: Dict[str, Any] = {}):
    # do a quick lookup and see if the project exists, if not, create it
    if project_id:
      project_id, project_name = self._get_name_id(project_id=project_id)
      if not project_id:
        raise Exception(f"Project with id {project_id} not found, please create a new one with project_name")
    else:
      project_id, project_name = self._get_name_id(project_name=project_name)
      if not project_id:
        logger.info(f"Project '{project_name}' not found, create one from dashboard.")

    # this is the config value that is used to store data on the plaform, user cannot be allowed to have
    # like a full access to config values
    log_config: Dict[str, Any] = {
      "user_config": config
    }

    # check if the current folder from where this code is being executed has a .git folder
    # NOTE: in case of NBX-Jobs the current folder ("./") is expected to contain git by default
    log_config["git"] = None
    if os.path.exists(".git"):
      log_config["git"] = get_git_details("./")

    # continue as before
    self._agent_details = AgentDetails(
      nbx_job_id = self._nbx_job_id or "jj_guvernr",
      nbx_run_id = self._nbx_run_id or "fake_run",
    )
    run_details = self.lmao.init_run(
      _InitRunRequest = InitRunRequest(
        agent_details=self._agent_details,
        created_at = get_timestamp(),
        project_name = project_name,
        project_id = project_id,
        config = dumps(log_config),
      )
    )

    if not run_details:
      # TODO: Make a custom exception of this
      raise Exception("Server Side exception has occurred, Check the log for details")

    if not project_id:
      project_id, project_name = self._get_name_id(project_name = project_name)

    self.project_name = project_name
    self.project_id = project_id
    self.config = config

    # TODO: change f-string here when refactoring `run_id` to `exp_id`
    # logger.info(f"Assigned run_id: {run_details.run_id}")
    self.run = run_details
    logger.info(f"Created a new LMAO run")
    logger.info(f"    id: {run_details.run_id}")
    logger.info(f"  link: https://app.nimblebox.ai/workspace/{self.workspace_id}/monitoring/{self.project_id}/{run_details.run_id}")

    # now initialize the relic
    if self.save_to_relic:
      # The relic will be the project id
      self.relic = RelicsNBX("experiments", self.workspace_id, create = True)
      logger.info(f"Will store everything in folder: {self.experiment_prefix}")

    # system metrics monitoring, by default is enabled optionally turn it off
    if self.enable_system_monitoring:
      self.system_monitoring = SystemMetricsLogger(self)
      self.system_monitoring.start()

  @property
  def experiment_prefix(self):
    prefix = f"{self.project_id}/{self.run.run_id}"
    if self._nbx_job_id is not None:
      prefix += f"_{self._nbx_job_id}@{self._nbx_run_id}"
    prefix += "/"
    return prefix

  """The functions below are the ones supposed to be used."""

  def log(self, y: Dict[str, Union[int, float, str]], step = None, *, log_type: str = RunLog.LogType.USER):
    """Log a single level dictionary to the platform at any given step. This function does not really care about the
    information that is being logged, it just logs it to the platform."""
    if self.completed:
      raise Exception("Run already completed, cannot log more data!")

    if self._total_logged_elements % DEBUG_LOG_EVERY == 0:
      logger.debug(f"Logging: {y.keys()} | {self._total_logged_elements}")

    if self._total_logged_elements % INFO_LOG_EVERY == 0:
      logger.info(f"Logging: {y.keys()} | {self._total_logged_elements}")

    step = step if step is not None else get_timestamp()
    if step < 0:
      raise Exception("Step must be <= 0")
    run_log = RunLog(run_id = self.run.run_id, log_type=log_type)
    for k,v in y.items():
      record = get_record(k, v)
      record.step = step
      run_log.data.append(record)

    ack = self.lmao.on_log(_RunLog = run_log)
    if not ack.success:
      logger.error("  >> Server Error")
      for l in ack.message.splitlines():
        logger.error("  " + l)
      raise Exception("Server Error")

    self._total_logged_elements += 1

  def save_file(self, *files: List[str]):
    """
    Register a file save. User should be aware of some structures that we follow for standardizing the data.
    All the experiments are going to be tracked under the following pattern:

    #. ``relic_name`` is going to be the experiment ID, so any changes to the name will not affect relic storage
    #. ``{experiment_id}(_{job_id}@{run_id})`` is the name of the folder which contains all the artifacts in the experiment.

    If relics is not enabled, this function will simply log to the LMAO DB.

    dk.save_file("foo.t", "/bar/", "baz.t", "/bar/roo/")
    """
    logger.info(f"Saving files: {files}")

    # manage all the complexity of getting the list of RelicFile
    all_files = []
    for folder_or_file in files:
      if os.path.isfile(folder_or_file):
        all_files.append(folder_or_file)
      elif os.path.isdir(folder_or_file):
        all_files.extend(U.get_files_in_folder(folder_or_file, self.username, self.workspace_id))
      else:
        raise Exception(f"File or Folder not found: {folder_or_file}")

    logger.debug(f"Storing {len(all_files)} files")
    if self.save_to_relic:
      logger.info(f"Uploading files to relic: {self.relic}")
      for f in all_files:
        print(f, f"{self.experiment_prefix}{f.strip('/')}")
        self.relic.put_to(f, f"{self.experiment_prefix}{f.strip('/')}")

    # log the files in the LMAO DB for sanity
    fl = FileList(run_id = self.run.run_id)
    fl.files.extend([File(relic_file = RelicFile(name = x)) for x in all_files])
    self.lmao.on_save(_FileList = fl)

  def end(self):
    """End thr run to declare it complete. This is more of a convinience function than anything else. For example when you
    are monitoring a live API you may never know when the experiment is complete. This function locks the experiment name
    so it can't be modified."""
    if self.completed:
      logger.error("Run already completed, cannot end it again!")
      return None

    logger.info("Ending run")
    ack = self.lmao.on_train_end(_Run = Run(run_id=self.run.run_id,))
    if not ack.success:
      logger.error("  >> Server Error")
      for l in ack.message.splitlines():
        logger.error("  " + l)
      raise Exception("Server Error")
    self.completed = True
    self.system_monitoring.stop()

"""
Utility functions below.
"""

def get_record(k: str, v: Union[int, float, str]) -> Record:
  """Function to create a Record protobuf object from a key and value."""
  _tv = type(v)
  assert _tv in [int, float, str], f"[key = {k}] '{_tv}' is not a valid type"
  _vt = {
    int: Record.DataType.INTEGER,
    float: Record.DataType.FLOAT,
    str: Record.DataType.STRING,
  }[_tv]
  record = Record(key = k, value_type = _vt)
  if _tv == int:
    record.integer_data.append(v)
  elif _tv == float:
    record.float_data.append(v)
  elif _tv == str:
    record.string_data.append(v)
  return record

def get_git_details(folder):
  """If there is a .git folder in the folder, return some details for that."""
  repo = Repo(folder)

  # check for any unstaged files
  uncommited_files = {}
  diff = repo.index.diff(None)
  for f in diff:
    path = f.a_path or f.b_path # when new file is added, a_path is None
    uncommited_files[path] = f.change_type
  if uncommited_files:
    logger.warning(f"Uncommited files: {uncommited_files}")

  # get the remote url
  try:
    remote_url = repo.remote().url

    # clean the remote_url because it can sometimes contain the repo token as well.
    # this can become security hazard. so if you have an example, that is not suppported:
    #   go ahead, make a PR!
    if "github.com" in remote_url:
      remote_url = re.sub(r"ghp_\w+@", "", remote_url)
  except ValueError:
    remote_url = None

  # get the size of the repository
  size = None
  for line in repo.git.count_objects("-v").splitlines():
    if line.startswith("size:"):
      size = int(line[len("size:") :].strip())
  if size > (1 << 30):
    logger.warning(f"Repository size over 1GB, you might want to work on it")

  return {
    "remote_url": remote_url,
    "branch": repo.active_branch.name,
    "commit": repo.head.commit.hexsha,
    "uncommited_files": uncommited_files,
    "untracked_files": repo.untracked_files,
    "size": size,
  }

def get_timestamp():
  """Get the current timestamp."""
  return int(datetime.now(timezone.utc).timestamp())



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""
import builtins
import google.protobuf.descriptor
import google.protobuf.message
import google.protobuf.timestamp_pb2
import typing
import typing_extensions

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

class Echo(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor
    SERVER_TIME_FIELD_NUMBER: builtins.int
    MESSAGE_FIELD_NUMBER: builtins.int
    ERROR_FIELD_NUMBER: builtins.int
    TRACEBACK_FIELD_NUMBER: builtins.int
    RPC_NAME_FIELD_NUMBER: builtins.int
    PROTO_DATA_FIELD_NUMBER: builtins.int
    BASE64_STRING_FIELD_NUMBER: builtins.int
    @property
    def server_time(self) -> google.protobuf.timestamp_pb2.Timestamp: ...
    message: typing.Text
    error: typing.Text
    traceback: typing.Text
    rpc_name: typing.Text
    proto_data: typing.Text
    """proto_data is the jsonified message
    on server side loads(proto_data) -> dict_to_message
    on client side proto_data = message_to_json(client_message)
    """

    base64_string: typing.Text
    """protobuf serilaised to the a base64 string"""

    def __init__(self,
        *,
        server_time: typing.Optional[google.protobuf.timestamp_pb2.Timestamp] = ...,
        message: typing.Text = ...,
        error: typing.Text = ...,
        traceback: typing.Text = ...,
        rpc_name: typing.Text = ...,
        proto_data: typing.Text = ...,
        base64_string: typing.Text = ...,
        ) -> None: ...
    def HasField(self, field_name: typing_extensions.Literal["data",b"data","error",b"error","message",b"message","server_time",b"server_time","traceback",b"traceback"]) -> builtins.bool: ...
    def ClearField(self, field_name: typing_extensions.Literal["base64_string",b"base64_string","data",b"data","error",b"error","message",b"message","proto_data",b"proto_data","rpc_name",b"rpc_name","server_time",b"server_time","traceback",b"traceback"]) -> None: ...
    def WhichOneof(self, oneof_group: typing_extensions.Literal["data",b"data"]) -> typing.Optional[typing_extensions.Literal["message","error","traceback"]]: ...
global___Echo = Echo



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nbox',
 'nbox.framework',
 'nbox.hyperloop',
 'nbox.lib',
 'nbox.nbxlib',
 'nbox.observability',
 'nbox.relics',
 'nbox.sub_utils',
 'nbox.sublime',
 'nbox.sublime._yql',
 'nbox.sublime.proto']

package_data = \
{'': ['*'], 'nbox': ['assets/*'], 'nbox.framework': ['protos/*']}

install_requires = \
['GitPython==3.1.27',
 'Jinja2==3.0.3',
 'cloudpickle==2.1.0',
 'gputil==1.4.0',
 'grpcio==1.43.0',
 'mypy-protobuf==3.2.0',
 'protobuf==3.20.1',
 'psutil==5.9.0',
 'python-json-logger==2.0.2',
 'randomname>=0.1.3,<0.2.0',
 'requests>=2.25.1,<3.0.0',
 'tabulate==0.8.9',
 'tqdm==4.64.0']

extras_require = \
{'serving': ['fastapi>=0.78.0,<0.79.0', 'uvicorn>=0.18.2,<0.19.0']}

entry_points = \
{'console_scripts': ['nbx = nbox.cli:main']}

setup_kwargs = {
    'name': 'nbox',
    'version': '0.9.42rc12',
    'description': 'ML Inference 🥶',
    'long_description': '<a href="https://nimblebox.ai/" target="_blank"><img src="./assets/built_at_nbx.svg" align="right"></a>\n[![PyPI - Python\nVersion](https://img.shields.io/badge/python-3.6%20%7C%203.7%20%7C%203.8%20%7C%203.9-blue)](https://pypi.org/project/nbox/)\n[![Downloads](https://pepy.tech/badge/nbox)](https://pepy.tech/project/nbox)\n![GitHub](https://img.shields.io/badge/license-Apache--2.0-blueviolet)\n\n## 🧐 What is Nbox?\n\n`nbox` provides first class support API for all NimbleBox.ai infrastructure (NBX-Build, Jobs, Deploy) and services (NBX-Workspaces) components. Write jobs using `nbox.Operators`\n\n# 🤷Why NimbleBox\n\n- Write and execute code in Python\n- Document your code that supports mathematical equations\n- Create/Upload/Share notebooks\n- Import notebooks from your local machine\n- Import/Publish notebooks from/to GitHub\n- Import external datasets (e.g. from Kaggle)\n- Integrate PyTorch, TensorFlow, Keras, OpenCV\n- Share your projects\n- Collaborate with your team\n\n# 🎚 Features\n\n### 🏗️ Freedom To Build\n![Build Landing Page (2)](https://user-images.githubusercontent.com/89596037/181773716-ba63f167-af0d-48aa-921a-02e13238c0f2.gif)\n\n\n### 🦾 Automate with Ease\n![Jobs Landing Page](https://user-images.githubusercontent.com/89596037/181774553-99120354-72f5-4064-9216-4f8a5aa050be.gif)\n\n\n### 🚀 Intuitive Dashboard\n![Deploy Landing Page](https://user-images.githubusercontent.com/89596037/181775468-cc342a30-d87e-4576-8bdd-8ffdd75ff759.gif)\n\n\n# 🏁 Get Started\n\n\n**Install the package from pipy:**\n\n```pip install nbox```\n\n\nFor convinience you should add nbox to your path by setting up an alias. Throughout the rest of the documentation we will be using nbx as the CLI:\n\n```# go to your .bashrc or .zshrc and add\nalias nbx="python3 -m nbox"\n```\n\n\n\nWhen loading nbox for the first time, it will prompt you the username and password and create a secrets file at ```~/.nbx/secrets.json. ``` This file then contains all the information that you don’t have to fetch manually again.\n\n\n## APIs\n\nOur APIs are deep, user functions are kept to minimum and most relavant. This documentation contains the full spec of everything, but here’s all the APIs you need to know:\n\n```\nnbox\n├── Model          # Framework agnostic Model\n│   ├── __call__\n│   ├── deploy\n│   ├── train_on_instance (WIP)\n│   └── train_on_jobs (WIP)\n├── Operators      # How jobs are combinations of operators\n│   ├── __call__\n│   └── deploy\n├── Jobs           # For controlling all your jobs\n│   ├── logs       # stream logs right on your terminal\n│   └── trigger    # manually trigger a job\n└── Instance\n   ├── __call__    # Run any command on the instance\n   └── mv (WIP)    # Move files to and from NBX-Build\n```\n\n## NBX-Jobs and NBX-Deploy\n\nCreate a simple folder:\n\n```\nnbx jobs new moonshot\n```\n\n### Run any Batch Process\n\nAdd the following code to `foobar.py`:\n\n```python\nfrom nbox import operator, Operator\nfrom nbox.lib.shell import ShellCommand\n\n# define your function and wrap it as an operator\n@operator()\ndef foo(x: Dict):\n  return "bar"\n\n# or use OOPs like torch, etc.\nclass MyJob(Operator):\n  def __init__(self):\n    super().__init__()\n    self.foo = foo\n    self.shell = ShellCommand("echo \'hello {text}\'")\n  \n  def forward(self):\n    text = self.foo()\n    self.shell(text = text)\n```\n\n# How to get help?\n\nJoin our [discord](https://discord.gg/qYZHxMaCsE) and someone from our community or engineering team will respond!\n\n## 🔖Read our [Blog](https://nimblebox.ai/blog).\n\n\n# 🧩 License\n\nThe code in thist repo is licensed as [Apache License 2.0](./LICENSE). Please check for individual repositories for licenses.\n',
    'author': 'NBX Research',
    'author_email': 'research@nimblebox.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/NimbleBoxAI/nbox',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
